/* ui.c
	Simple user interface for the GameBoy
	Michael Hope, 1998
	michaelh@earthling.net
	Distrubuted under the Artistic License - see www.opensource.org
*/

#include "ui.h"
#include "font.h"
#include <gb.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/* Definitions describing which tiles in font_draw should be used
	for drawing various objects */
/* Convention:
	More than one line object:
		_TL	- top left
		_TM	- top middle
		_TR	- top right
		_L	- middle left
		_R	- middle right
		_BL, _BM, _BR - as for _T*
*/
/* Tiles for drawing a button */
#define TBUT		4
#define TBUT_TL		(TBUT+0)
#define TBUT_TM		(TBUT+1)
#define TBUT_TR		(TBUT+2)
#define TBUT_L		(TBUT+3)
#define TBUT_R		(TBUT+4)
#define TBUT_BL		(TBUT+5)
#define TBUT_BM		(TBUT+6)
#define TBUT_BR		(TBUT+7)

/* ..for drawing a box (window border) */
#define TBOX		20
#define TBOX_TL		(TBOX+0)
#define TBOX_TM		(TBUT+6)
#define TBOX_TR		(TBOX+1)
#define TBOX_L		(TBUT+4)
#define TBOX_R		(TBUT+3)
#define TBOX_BL		(TBOX+2)
#define TBOX_BM		(TBUT+1)
#define TBOX_BR		(TBOX+3)

/* Fuzz for drawing a title bar for a screen or window */
#define TFUZZ		24
#define TFUZZ_L		(TFUZZ+0)
#define TFUZZ_M		(TFUZZ+1)
#define TFUZZ_R		(TFUZZ+2)

/* For drawing a slider */
#define TSLIDER		27
#define TSLIDER_DOT	(TSLIDER+0)
#define TSLIDER_L	(TSLIDER+1)
#define TSLIDER_M	(TSLIDER+2)
#define TSLIDER_R	(TSLIDER+3)

/*#define HACK*/	/* Precisely matched char[] lengths to get around a problem with .blkb */

/* Current position of the pointer */
UBYTE point_x, point_y;

/* Handler definitions - handlers for each of the main objects */
mhandler screen_handler = {
	screen_paint,
	find_focus
};

mhandler button_handler = {
	button_paint,
	button_but_down
};

mhandler frame_handler = {
	frame_paint,
	find_focus
};

mhandler window_handler = {
	window_paint,
	find_focus
};

mhandler static_handler = {
	static_paint,
	NULL
};

mhandler spinner_handler = {
	spinner_paint,
	spinner_but_down
};

mhandler slider_handler = {
	slider_paint,
	slider_but_down
};

mhandler image_handler = {
	image_paint,
	NULL
};

/* Font handles.  Initalised by ui_init() */
font_t hfont_draw, hfont_text, hfont_text_inv;

/* but_down rotuines - routines that handle what happens when the user
   presses a button down over an object.
   Any object that has children uses find_focus to see which child should
     take the click
   Any object that cant be clicked on has NULL in its but_down handler field 
   Any other object must provide it's own _but_down routine
*/

int button_paint_sel(pmobject This, UBYTE x, UBYTE y);

/* Handler for a click over a button */	
int button_but_down(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay)
{
	UBYTE key;
	button_paint_sel(This, px, py);
	key=J_A;
	while (key&J_A) {
		key=joypad();
	}
	button_paint(This, px, py);
	return ((pmbutton)This)->id;
}

/* Handler for a click over a spinner */
int spinner_but_down(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay)
{
	UBYTE key, i;
	pmspinner spin;

	spin=(pmspinner)This;
	key=J_A;
	while (key&J_A) {
		key=joypad();
		if (key&(J_UP)) {
			spin->current--;
			if (spin->current<spin->min)
				spin->current=spin->max;
		}
		if (key&(J_DOWN)) {
			spin->current++;
			if (spin->current>spin->max)
				spin->current=spin->min;
		}
		if ((key&J_UP)||(key&J_DOWN)) {
			This->handler->paint(This, px, py);
			for (i=0; i<12; i++)
				wait_vbl_done();
		}
	}
	return 0;
}

/* Handler for a click over a slider */
int slider_but_down(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay)
{
	UBYTE key, i;
	pmslider slide;

	slide=(pmslider)This;
	key=J_A;
	while (key&J_A) {
		key=joypad();
		if (key&(J_LEFT)) {
			slide->current-=slide->step;
			if (slide->current<slide->min)
				slide->current=slide->min;
		}
		if (key&(J_RIGHT)) {
			slide->current+=slide->step;
			if (slide->current>slide->max)
				slide->current=slide->max;
		}
		if ((key&J_LEFT)||(key&J_RIGHT)) {
			This->handler->paint(This, px, py);
			if (slide->change_hook != NULL)
				slide->change_hook(This);
			for (i=0; i<12; i++)
				wait_vbl_done();
		}
	}
	return 0;
}

/* Handler for any object with children.  Finds which child was clicked
   on and passes control to it's handler.
   Returns 0 if the click doesnt cover anything clickable
*/
int find_focus(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay)
{
	pmobject walk;

	if (This->focus==NULL) {
		/* Scan all the children to find who owns this click */
		walk=This->children;
		while (walk) {
			if ((walk->left<=x)&&(walk->top<=y)&&
			    ((walk->left+walk->width)>x)&&((walk->top+walk->height)>y)) {
				if (walk->handler->but_down!=NULL) {
					return walk->handler->but_down(walk, px+This->left, py+This->top, (x-walk->left), (y-walk->top), ax, ay);
				}
			}
			walk=walk->next;
		}
		return 0;
	}
	else {
		walk=This->focus;
		return walk->handler->but_down(walk, px+This->left, py+This->top, (x-walk->left), (y-walk->top), ax, ay);
	}
}
	
int add_to_children( pmobject parent, pmobject child )
{
	pmobject walk;

	/* Add to the tail of the list */
	if (parent) {
		if (parent->children==NULL) {
			parent->children=child;
		}
		else {
			walk=parent->children;
			while (walk) {
				if (walk->next==NULL) {
					walk->next=child;
					child->next=NULL;
					walk=NULL;
				}
				else
					walk=walk->next;
			}
		}
		return 0;
	}
	return 1;
}

/* Paint routines - functions which draw an object and all of it's children */
int paint_children( pmobject object, UBYTE x, UBYTE y )
{
	pmobject walk = object->children;

	if (object->focus==NULL) {
		while (walk) {
			walk->handler->paint(walk, x, y);
			walk=walk->next;
		}
	}
	else {
		object->focus->handler->paint(object->focus, x, y);
	}
	return 0;
}

/* Paint an image */
int image_paint(pmobject This, UBYTE x, UBYTE y)
{
	pmimage img;
	UBYTE cell, row, col;

	cell = 0;
	row = 0;
	font_set(((pmimage)This)->hfont);

	while (row<This->height) {
		mgotoxy( This->left+x, This->top+y+row );
		for (col=0; col<This->width; col++) {
			mput_char(cell++);
		}
		row++;
	}
	/* An image cant have any children */
	return 0;
}

/* Paint a spinner */
int spinner_paint(pmobject This, UBYTE x, UBYTE y)
{
	pmspinner spin;
	char	buffer[10];

	spin=(pmspinner)This;

	font_set(hfont_text);
	sprintf(buffer, "%ld", spin->current);
	mgotoxy(spin->left+x, spin->top+y);
	if ((spin->max>100)&&(spin->current<100))
		mput_char(' ');
	if ((spin->max>10)&&(spin->current<10))
		mput_char(' ');
	mprint_string(buffer);
	if (spin->unit!=0) {
		mput_char(spin->unit);
	}
	return 0;
}

/* Paint a static */
int static_paint(pmobject This, UBYTE x, UBYTE y)
{
	pmstatic stat;
	stat=(pmstatic)This;
	font_set(hfont_text);
	switch (stat->just) {
		case JUST_RIGHT:
			x+=stat->width-strlen(stat->label);
			break;
		case JUST_CENTRE:
			x+=((stat->width-strlen(stat->label))/2);
			break;
		default:
			break;
	}
	mgotoxy(stat->left+x, stat->top+y);
	mprint_string(stat->label);
	/* static cant have any children */
	return 0;
}

/* Paint a button */
int button_paint(pmobject This, UBYTE x, UBYTE y)
{
	pmbutton button;
	int i, j;

	button=(pmbutton)This;
	mgotoxy(button->left+x, button->top+y );
	font_set(hfont_draw);
	mput_char(TBUT_TL);
	for (i=2; i<button->width; i++) {
		mput_char(TBUT_TM);
	}
	mput_char(TBUT_TR);
	mgotoxy(button->left+x, button->top+y+1);
	mput_char(TBUT_L);
	font_set(hfont_text);
	i = (button->width - strlen(button->label))/2;
	j = 1;
	while (j<i) {
		mput_char(TCLEAR);
		j++;
	}
	mprint_string(button->label);
	j+=strlen(button->label)+1;
	while (j<button->width) {
		mput_char(TCLEAR);
		j++;
	}
	font_set(hfont_draw);
	mput_char(TBUT_R);
	mgotoxy(button->left+x, button->top+y+2 );
	mput_char(TBUT_BL);
	for (i=2; i<button->width; i++) {
		mput_char(TBUT_BM);
	}
	mput_char(TBUT_BR);
	
	paint_children(This,x+button->left,y+button->top);
	return 0;
}

/* Paint a slider */
int slider_paint(pmobject This, UBYTE x, UBYTE y)
{
	UBYTE pos, slider_width;
	pmslider slider;
	char buffer[5];
	
	slider = (pmslider)This;
	slider_width = slider->width-slider->val_width;

	sprintf(buffer, "%ld", slider->current);
	buffer[slider->val_width+1] = '\0';
	
	pos = slider_width*(slider->current - slider->min) / (slider->max - slider->min);
	mgotoxy( slider->left+x, slider->top+y);
	if (slider->current==slider->max)
		pos--;		/* Take care of rounding errors */
	font_set(hfont_draw);
	if (pos>0) {
		mput_char(TSLIDER_L);
		print_repeat( TSLIDER_M, pos-1);
	}
	mput_char(TSLIDER_DOT);
	if (pos<(slider_width-1)) {
		print_repeat( TSLIDER_M, slider_width - 2 - pos);
		mput_char( TSLIDER_R);
	}
	font_set(hfont_text);
	print_repeat(' ', slider->val_width - strlen(buffer));
	mprint_string(buffer);
	paint_children(This,x+slider->left,y+slider->top);
	return 0;
}

/* Paint a slected (black) button 
   Used only by button_but_down 
*/
int button_paint_sel(pmobject This, UBYTE x, UBYTE y)
{
	pmbutton button;
	int i, j;

	button=(pmbutton)This;
	mgotoxy(button->left+x, button->top+y );
	font_set(hfont_draw);
	mput_char(TBUT_TL+8);
	for (i=2; i<button->width; i++) {
		mput_char(TBUT_TM+8);
	}
	mput_char(TBUT_TR+8);
	mgotoxy(button->left+x, button->top+y+1);
	mput_char(TBUT_L+8);
	i = (button->width - strlen(button->label))/2;
	j = 1;
	while (j<i) {
		mput_char(TSOLID);
		j++;
	}
	font_set(hfont_text_inv);
	mprint_string(button->label);
	font_set(hfont_draw);
	j+=strlen(button->label)+1;
	while (j<button->width) {
		mput_char(TSOLID);
		j++;
	}
	font_set(hfont_draw);
	mput_char(TBUT_R+8);
	mgotoxy(button->left+x, button->top+y+2 );
	mput_char(TBUT_BL+8);
	for (i=2; i<button->width; i++) {
		mput_char(TBUT_BM+8);
	}
	mput_char(TBUT_BR+8);
	
	paint_children(This,x+button->left,y+button->top);
	return 0;
}

/* Paint a screen by painting its title and all of it's children */
int screen_paint(pmobject This, UBYTE x, UBYTE y)
{
	pmscreen screen;
	screen=(pmscreen)This;
	draw_title_bar(0,0,20,screen->label, 0 );
	paint_children( This, x, y );
	return 0;
}

/* Draw a frame by drawing all of it's children */
int frame_paint(pmobject This, UBYTE x, UBYTE y)
{
	pmframe frame;
	frame=(pmframe)This;
	paint_children( This, frame->left+x, frame->top+y );
	return 0;
}

/* Draw a window and paint it's children */
int window_paint(pmobject This, UBYTE x, UBYTE y)
{
	pmwindow window;
	window=(pmwindow)This;
	draw_window(window->left+x, window->top+y, window->width, window->height, window->label );
	paint_children( This, window->left+x, window->top+y );
	return 0;
}

/* Erase the rectangle that covers the object 'This'
   Used to 'pop down' a window or button before re-drawing the objects
   below it 
*/
int erase_rect(pmobject This, UBYTE skip)
{
	UBYTE y, count;
	y = This->top+skip;
	count = This->height;
	font_set(hfont_draw);
	while (count--) {
		mgotoxy(This->left, y++);
		print_repeat(TCLEAR, This->width);
	}
	return 0;
}

/* Erase the whole screen.  Not sure why you'd want to */
int screen_clear(pmscreen screen)
{
	return erase_rect((pmobject)screen, 1);
}
		
/* Initialise a screen object
   Used to init a screen at the very start of a main loop */
pmscreen screen_init(pmscreen screen, UBYTE top, UBYTE left, UBYTE width, UBYTE height, char *title)
{
	if (screen==NULL) {
		screen=calloc(sizeof(mscreen),1);
	}
	if (screen) {
		strcpy( screen->label, title);
		screen->top=top;
		screen->left=left;
		screen->width=width;
		screen->height=height;
		screen->next=NULL;
		screen->children=NULL;
		screen->parent=NULL;
		screen->handler=&screen_handler;
		screen->focus=NULL;
	}
	return screen;
}

/* Initalise a frame */
pmframe frame_init(pmobject parent, UBYTE left, UBYTE top, UBYTE width, UBYTE height)
{
	pmframe frame;

	frame=calloc(sizeof(mframe),1);
	if (frame) {
		frame->top=top;
		frame->left=left;
		frame->width=width;
		frame->height=height;
		frame->next=NULL;
		frame->children=NULL;
		frame->parent=parent;
		add_to_children( parent, (pmobject)frame );
		frame->handler=&frame_handler;
	}
	return frame;
}

/* Standard pointer sprite */
static unsigned char point_sprite[] = 
	{ 0xFF,0x00,0xFF,0x7E,0xFF,0x7C,0xFE,0x78,0xFF,0x7C,0xFF,0x6E,0xFF,0x46,0xEF,0x00 };

/* Initalise the pointer */
int point_init(void)
{
	/* Initialize sprite palette */
	OBP1_REG = 0xE0U;
	set_sprite_data(0, 1, point_sprite);
	set_sprite_tile(0,0);
	set_sprite_prop(0,0x10);
	move_sprite(0,point_x+8,point_y+16);
	SHOW_SPRITES;
	return 0;
}

/* Initalise the whole ui system.
   font_draw is the font you want to use for drawing the objects
   font_text is the font you want to use for all the text
     Both a standard and inverted version is used, so make sure
     that sufficient tiles are avaliable to fit the text font twice
     and the drawing font _and_ any image fonts you want
*/
int ui_init(void *font_draw, void *font_text)
{	
	font_mode |= FONT_MODE_NOSCROLL;
	hfont_draw = font_load( font_draw );
	hfont_text = font_load( font_text );
	font_bg_colour = 3;
	hfont_text_inv = font_load( font_text );
	point_x=0;
	point_y=0;
	point_init();
	return 0;
}

/* Drawing routines - primitaves that are called by the paint routines */

/* Draw a window (box with title bar) */
int draw_window( UBYTE left, UBYTE top, UBYTE width, UBYTE height, char *string )
{
	draw_box(left, top, width, height, 1 );
	draw_title_bar(left,top, width, string, 1);
	return 0;
}

/* Draw a title bar with text centre justified */
int draw_title_bar( UBYTE left, UBYTE top, UBYTE width, char *string, UBYTE mode )
{
	UBYTE i;
	UBYTE len, center, span;

	len = strlen( string );
	center = width/2;
	span = (width-len)/2;
	
	mgotoxy( left, top );
	i=0;
	font_set( hfont_draw );
	if (mode==1) {
		mput_char( TFUZZ_L );
		i++;
	}
	for (; i<span; i++ ) {
		mput_char( TFUZZ_M );
	}
	font_set( hfont_text );
	mprint_string(string);
	i+=len;
	font_set( hfont_draw );
	for (; i<width-1; i++ ) {
		mput_char( TFUZZ_M );
	}
	if (mode==1) {
		mput_char( TFUZZ_R );
	}
	else
		mput_char( TFUZZ_M );
	return 0;
}

/* Draw a box.  If mode is 1 then the box has a top */
int draw_box( UBYTE left, UBYTE top, UBYTE width, UBYTE height, UBYTE mode )
{
	UBYTE i;
	
	font_set( hfont_draw );
	if (mode==1) {
		mgotoxy( left, top );
		mput_char( TBOX_TL );
		print_repeat( TBOX_TM, width-2 );
		mgotoxy( left+width-1, top );
		mput_char( TBOX_TR );
	}
	for (i=1; i<height-1; i++) {
		mgotoxy( left, i+top );
		mput_char(TBOX_L);
		print_repeat( TCLEAR, width-2);
		mput_char(TBOX_R);
	}
	mgotoxy( left, top+height-1 );
	mput_char( TBOX_BL );
	print_repeat(TBOX_BM, width-2);
	mput_char( TBOX_BR );
	return 0;
}

/* Pointer handler.
   Simple pointer move/button check routine 
   Uses the globals point_x and point_y to store the
   current/previous position of the pointer
*/
void pointer_loop(UBYTE mask, UBYTE logic)
{
	UBYTE key;
	UBYTE slowdown;
	UBYTE done;
	done=0;
	slowdown = 15;

	while (!done) {
		key = joypad();
		if(key & (J_UP|J_DOWN|J_LEFT|J_RIGHT)) {
			if (key & J_UP && point_y > 0)
				point_y--;
			if(key & J_DOWN && point_y < 143)
				point_y++;
			if(key & J_LEFT && point_x > 0)
				point_x--;
			if(key & J_RIGHT && point_x < 159)
				point_x++;
			move_sprite(0,point_x+8,point_y+16);
			while(slowdown && key == joypad()) {
				wait_vbl_done();
				slowdown--;
			}
			slowdown = 1;
		} 
		else
			slowdown = 10;
		if (logic) {
			/* 1 = down */
			done=key&mask;
		}
		else {
			done=!(key&mask);
		}
	}
}
