/* ui.h
	Simple user interface for the GameBoy
	Michael Hope, 1998
	michaelh@earthling.net
	Distrubuted under the Artistic License - see www.opensource.org
*/
#ifndef __UI_H
#define __UI_H

#include "font.h"
#include <gb.h>

/* Justification types for statics */
#define JUST_LEFT	0
#define JUST_RIGHT	1
#define JUST_CENTRE	2

/* Maximum label length.  As the font code doesnt
   support newlines, 20 is reasonable */
#define	LABEL_LEN	20

/* Standard tiles in font_draw, which give a blank, light grey, dark grey,
   and solid tile */
#define TCLEAR		0
#define TGREY33		1
#define TGREY66		2
#define TSOLID		3

typedef struct sobject mobject;
typedef struct sobject *pmobject;

/* A handler is a structure which contains pointers to the functions
   which can handle events on this object.  This is used to make C mildly
   object-orientated
   The 'paint' function knows how to draw (paint) this object on the screen
	'x' is the left co-ordinate of the parent
	'y' is the top co-ordinate of the parent
   The 'but_down' function knows how to handle a click over the object
	'px' is the absolute left co-ordinate of the parent in cells
	'py' is the absolute top co-ordinate of the parent in cells
	'x' is the left co-ordinate with respect to the parent in cells
	'y' is the top co-ordinate with respect to the parent in cells
	'ax' is the absolute x co-ordinate in pixels
	'ay' is the absolute y co-ordinate in pixels
	px and py are used for drawing.  x and y are used for finding which
	child should be clicked.  ax and ay are there for completeness.
*/
typedef struct shandler mhandler;
typedef struct shandler *pmhandler;

struct shandler {
	int (*paint)(pmobject This, UBYTE x, UBYTE y);
	int (*but_down)(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay);
};

/*
Standard object fields:

handler		handler for the object
children	First in the objects children
parent		Parent of this object
next		Next in this list of children
left, top, width, height	Co-ordinates of bounding box
*/
/* STD_DATA is the data which is common to _all_ objects */
#define STD_DATA	pmhandler handler; \
			pmobject children; \
			pmobject parent; \
			pmobject focus; \
			pmobject next; \
			UBYTE left, top, width, height;

struct sobject {
	STD_DATA
};

/* A screen is the top level object.  It must be allocated in RAM
   as you must be able to add children to it.
   A screen consists of a title bar with the text 'label' centre justified.
   If you want to give focus to a popped-up window, set the 'focus' field
   to point to the appropriate window.  Same for any 3rd or greater level windows.
*/
typedef struct sscreen mscreen;
typedef struct sscreen *pmscreen;

struct sscreen {
	STD_DATA
	char	label[LABEL_LEN];
	
};

/* A button is a 3 line high object with the centre justified text 'label'
   which is surrounded by a border.
   Buttons can be clicked.  When they are clicked, they return the event
   code 'id'
*/
typedef struct sbutton mbutton;
typedef struct sbutton *pmbutton;

struct sbutton {
	STD_DATA
	char	label[LABEL_LEN];
	int	id;
};

/* A frame is a container for many other objects.  A frame doesnt draw
   anything on the screen directly.
   The advantage of a frame is that you can place many objects - for example
   a slider and it's label - within it and then move them as a group instead
   of having to shift them individually 
*/
typedef struct sframe mframe;
typedef struct sframe *pmframe;

struct sframe {
	STD_DATA
};

/* A window is a a box with a title bar and the text 'label' centre justified.
   A window can be the focus of a screen so that the window is the only thing
   that can be clicked upon - see above 
*/
typedef struct swindow mwindow;
typedef struct swindow *pmwindow;

struct swindow {
	STD_DATA
	char	label[LABEL_LEN];
};

/* A static is a static text string, consisting of the text 'label' justified
   according to 'just' inside it's bounding box.  just should be one of JUST_*
   above
*/
typedef struct sstatic mstatic;
typedef struct sstatic *pmstatic;

struct sstatic { 
	STD_DATA
	char	label[LABEL_LEN];
	UBYTE	just;
};

/* A spinner is a number that can be changed by clicking and dragging
   over it.  The minimum, maximum, and step between values can be set using
   min, max, and step.  The inital value is in current.  unit is an optional
   character that is displayed after the value as a unit (eg 's' for seconds).
   The change_hook entry allows a function to be called whenever the value of
   the spinner changes.
*/
/* A spinner must exist in RAM */
typedef struct sspinner mspinner;
typedef struct sspinner *pmspinner;

struct sspinner {
	STD_DATA
	WORD	current;
	WORD	min;
	WORD	max;
	WORD	step;
	char	unit;
	void	(*change_hook)(pmobject This);
};

/* A slider is a one line high bar with a marker that represents a value.
   The minimum, maximum, and step between values can be set using
   min, max, and step.  The inital value is in current.
   The change_hook entry allows a function to be called whenever the value of
   the spinner changes.
*/
/* A slider must exist in RAM */
typedef struct sslider mslider;
typedef struct sslider *pmslider;

struct sslider {
	STD_DATA
	WORD	current;
	WORD	min;
	WORD	max;
	WORD	step;
	UBYTE	val_width;	/* Space to reserve for value */
	void	(*change_hook)(pmobject This);
};

/* A image is a rectangular array of font tiles that can be used to
   display a simple image.  font points to the already loaded font
   (using font_load) that contains the required tiles.  The bounding
   box is used to display the tiles in the right order.  For example,
   a 2 wide, 3 high, 6 cell image with width=2, height=3 would be drawn
   as (number = tile number)
	01
	23
	45
   A image with width=4, height=2 would be drawn as
	0123
	4567
*/
/* A image must be stored in RAM as the font entry needs to be updtaed at
   run time
*/
typedef struct simage mimage;
typedef struct simage *pmimage;

struct simage {
	STD_DATA
	font_t	hfont;
};

/* Handles of the fonts loaded during ui_init()
	hfont_draw is used to draw all of the objects
	hfont_text is used for the labels and text
	hfont_text_inv is an inverted copy of hfont_text used to show selection
*/
extern UWORD hfont_draw, hfont_text, hfont_text_inv;

/* The object handlers */
extern mhandler screen_handler;
extern mhandler button_handler;
extern mhandler frame_handler;
extern mhandler window_handler;
extern mhandler static_handler;
extern mhandler spinner_handler;
extern mhandler slider_handler;
extern mhandler image_handler;

int draw_title_bar( UBYTE left, UBYTE top, UBYTE right, char *string, UBYTE mode );
int draw_box( UBYTE left, UBYTE top, UBYTE width, UBYTE height, UBYTE mode );
int draw_window( UBYTE left, UBYTE top, UBYTE width, UBYTE height, char *string );
int screen_paint(pmobject This, UBYTE x, UBYTE y);
int button_paint(pmobject This, UBYTE x, UBYTE y);
int frame_paint(pmobject This, UBYTE x, UBYTE y);
int window_paint(pmobject This, UBYTE x, UBYTE y);
int static_paint(pmobject This, UBYTE x, UBYTE y);
int spinner_paint(pmobject This, UBYTE x, UBYTE y);
int slider_paint(pmobject This, UBYTE x, UBYTE y);
int image_paint(pmobject This, UBYTE x, UBYTE y);

int find_focus(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay);
int wait_but_up(pmobject This, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay);
int spinner_but_down(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay);
int button_but_down(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay);
int slider_but_down(pmobject This, UBYTE px, UBYTE py, UBYTE x, UBYTE y, UBYTE ax, UBYTE ay);
int screen_clear(pmscreen screen);
int erase_rect(pmobject This, UBYTE skip);

int ui_init(void *font_draw, void *font_text);
pmscreen screen_init(pmscreen screen, UBYTE top, UBYTE left, UBYTE width, UBYTE height, char *title);

extern UBYTE point_x, point_y;
void pointer_loop(UBYTE mask, UBYTE logic);

#endif /* __UI_H */
